# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides auxiliary math functions for PyAnsys Geometry."""

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.misc.accuracy import Accuracy
from ansys.geometry.core.typing import Real


@check_input_types
def get_two_circle_intersections(
    x0: Real, y0: Real, r0: Real, x1: Real, y1: Real, r1: Real
) -> tuple[tuple[Real, Real], tuple[Real, Real]] | None:
    """Get the intersection points of two circles.

    Parameters
    ----------
    x0 : Real
        x coordinate of the first circle.
    y0 : Real
        y coordinate of the first circle.
    r0 : Real
        Radius of the first circle.
    x1 : Real
        x coordinate of the second circle.
    y1 : Real
        y coordinate of the second circle.
    r1 : Real
        Radius of the second circle.

    Returns
    -------
    tuple[tuple[Real, Real], tuple[Real, Real]] | None
        Intersection points of the two circles if they intersect.
        The points are returned as ``((x3, y3), (x4, y4))``, where ``(x3, y3)`` and ``(x4, y4)``
        are the intersection points of the two circles. If the circles do not
        intersect, then ``None`` is returned.

    Notes
    -----
    This function is based on the following StackOverflow post:
    https://stackoverflow.com/questions/55816902/finding-the-intersection-of-two-circles

    That post is based on the following implementation:
    https://paulbourke.net/geometry/circlesphere/
    """
    # circle 1: (x0, y0), radius r0
    # circle 2: (x1, y1), radius r1
    d = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)

    # non-intersecting
    if d > r0 + r1:
        return None
    # one circle within other
    if d < abs(r0 - r1):
        return None
    # coincident circles
    if d == 0 and r0 == r1:
        return None
    else:
        a = (r0**2 - r1**2 + d**2) / (2 * d)
        # abs is used to ensure non-negative value - close to zero
        # conditions might occur if circles are nearly tangent
        h = np.sqrt(abs(r0**2 - a**2))
        x2 = x0 + a * (x1 - x0) / d
        y2 = y0 + a * (y1 - y0) / d

        dx = h * (y1 - y0) / d
        dy = h * (x1 - x0) / d
        x3 = x2 + dx
        y3 = y2 - dy

        x4 = x2 - dx
        y4 = y2 + dy

        return ((x3, y3), (x4, y4))


def intersect_interval(first_min, second_min, first_max, second_max) -> tuple[bool, Real, Real]:
    """Find the intersection of two intervals.

    Parameters
    ----------
    first_min : Real
        The minimum value of the first interval.
    second_min : Real
        The minimum value of the second interval.
    first_max : Real
        The maximum value of the first interval.
    second_max : Real
        The maximum value of the second interval.

    Returns
    -------
    tuple[bool, Real, Real]
        Tuple with a boolean to indicate whether the intervals intersect,
        the minimum value of the intersection interval,
        and the maximum value of the intersection interval.
        If they do not intersect, then the boolean is False and the other values are 0.
    """
    minimum = second_min
    if first_min > minimum:
        minimum = first_min

    maximum = second_max
    if first_max < maximum:
        maximum = first_max

    if minimum > maximum:
        if minimum - maximum > Accuracy.length_accuracy():
            return False, 0, 0

        maximum, minimum = minimum, maximum

    return True, minimum, maximum



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing an ellipse."""

from functools import cached_property

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity
from scipy.integrate import quad

from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Z
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.accuracy import Accuracy
from ansys.geometry.core.misc.measurements import Distance
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.curve_evaluation import CurveEvaluation
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
)
from ansys.geometry.core.typing import Real, RealSequence


class Ellipse(Curve):
    """Provides 3D ellipse representation.

    Parameters
    ----------
    origin : ~numpy.ndarray | RealSequence | Point3D
        Origin of the ellipse.
    major_radius : ~pint.Quantity | Distance | Real
        Major radius of the ellipse.
    minor_radius : ~pint.Quantity | Distance | Real
        Minor radius of the ellipse.
    reference : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
        X-axis direction.
    axis : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
        Z-axis direction.
    """

    @check_input_types
    def __init__(
        self,
        origin: np.ndarray | RealSequence | Point3D,
        major_radius: Quantity | Distance | Real,
        minor_radius: Quantity | Distance | Real,
        reference: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_X,
        axis: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_Z,
    ):
        """Initialize the ``Ellipse`` class."""
        self._origin = Point3D(origin) if not isinstance(origin, Point3D) else origin

        self._reference = (
            UnitVector3D(reference) if not isinstance(reference, UnitVector3D) else reference
        )
        self._axis = UnitVector3D(axis) if not isinstance(axis, UnitVector3D) else axis

        if not self._reference.is_perpendicular_to(self._axis):
            raise ValueError(
                "Ellipse reference (x-direction) and axis (z-direction) must be perpendicular."
            )

        self._major_radius = (
            major_radius if isinstance(major_radius, Distance) else Distance(major_radius)
        )
        self._minor_radius = (
            minor_radius if isinstance(minor_radius, Distance) else Distance(minor_radius)
        )

        if self._major_radius.value.m_as(self._major_radius.base_unit) <= 0:
            raise ValueError("Major radius must be a real positive value.")
        if self._minor_radius.value.m_as(self._minor_radius.base_unit) <= 0:
            raise ValueError("Minor radius must be a real positive value.")

        # Align both units if misaligned
        if self._major_radius.unit != self._minor_radius.unit:
            self._minor_radius.unit = self._major_radius.unit

        # Ensure that the major radius is equal or larger than the minor one
        if self._major_radius.value.m < self._minor_radius.value.m:
            raise ValueError("Major radius cannot be shorter than the minor radius.")

    @property
    def origin(self) -> Point3D:
        """Origin of the ellipse."""
        return self._origin

    @property
    def major_radius(self) -> Quantity:
        """Major radius of the ellipse."""
        return self._major_radius.value

    @property
    def minor_radius(self) -> Quantity:
        """Minor radius of the ellipse."""
        return self._minor_radius.value

    @property
    def dir_x(self) -> UnitVector3D:
        """X-direction of the ellipse."""
        return self._reference

    @property
    def dir_y(self) -> UnitVector3D:
        """Y-direction of the ellipse."""
        return self.dir_z.cross(self.dir_x)

    @property
    def dir_z(self) -> UnitVector3D:
        """Z-direction of the ellipse."""
        return self._axis

    @check_input_types
    def __eq__(self, other: "Ellipse") -> bool:
        """Equals operator for the ``Ellipse`` class."""
        return (
            self._origin == other._origin
            and self._major_radius == other._major_radius
            and self._minor_radius == other._minor_radius
            and self._reference == other._reference
            and self._axis == other._axis
        )

    def mirrored_copy(self) -> "Ellipse":
        """Create a mirrored copy of the ellipse along the y-axis.

        Returns
        -------
        Ellipse
            New ellipse that is a mirrored copy of the original ellipse.
        """
        return Ellipse(
            self.origin, self.major_radius, self.minor_radius, -self._reference, -self._axis
        )

    def evaluate(self, parameter: Real) -> "EllipseEvaluation":
        """Evaluate the ellipse at the given parameter.

        Parameters
        ----------
        parameter : Real
            Parameter to evaluate the ellipse at.

        Returns
        -------
        EllipseEvaluation
            Resulting evaluation.
        """
        return EllipseEvaluation(self, parameter)

    def project_point(self, point: Point3D) -> "EllipseEvaluation":
        """Project a point onto the ellipse and evaluate the ellipse.

        Parameters
        ----------
        point : Point3D
            Point to project onto the ellipse.

        Returns
        -------
        EllipseEvaluation
            Resulting evaluation.
        """
        origin_to_point = point - self.origin
        dir_in_plane = UnitVector3D.from_points(
            Point3D([0, 0, 0]), origin_to_point - ((origin_to_point * self.dir_z) * self.dir_z)
        )
        if dir_in_plane.is_zero:
            return EllipseEvaluation(self, 0)

        t = np.arctan2(
            self.dir_y.dot(dir_in_plane) * self.major_radius.m,
            self.dir_x.dot(dir_in_plane) * self.minor_radius.m,
        )
        return EllipseEvaluation(self, t)

    def is_coincident_ellipse(self, other: "Ellipse") -> bool:
        """Determine if this ellipse is coincident with another.

        Parameters
        ----------
        other : Ellipse
            Ellipse to determine coincidence with.

        Returns
        -------
        bool
            ``True`` if this ellipse is coincident with the other, ``False`` otherwise.
        """
        return (
            Accuracy.length_is_equal(self.major_radius.m, other.major_radius.m)
            and Accuracy.length_is_equal(self.minor_radius.m, other.minor_radius.m)
            and self.origin == other.origin
            and self.dir_z == other.dir_z
        )

    @property
    def eccentricity(self) -> Real:
        """Eccentricity of the ellipse."""
        ecc = (self.major_radius.m**2 - self.minor_radius.m**2) ** 0.5 / self.major_radius.m
        if ecc == 1:
            raise ValueError("The curve defined is a parabola and not an ellipse.")
        elif ecc > 1:
            raise ValueError("The curve defined is an hyperbola and not an ellipse.")
        return ecc

    @property
    def linear_eccentricity(self) -> Quantity:
        """Linear eccentricity of the ellipse.

        Notes
        -----
        The linear eccentricity is the distance from the center to the focus.
        """
        return (self.major_radius**2 - self.minor_radius**2) ** 0.5

    @property
    def semi_latus_rectum(self) -> Quantity:
        """Semi-latus rectum of the ellipse."""
        return self.minor_radius**2 / self.major_radius

    @property
    def perimeter(self) -> Quantity:
        """Perimeter of the ellipse."""

        def integrand(theta, ecc):
            return np.sqrt(1 - (ecc * np.sin(theta)) ** 2)

        integral, _ = quad(integrand, 0, np.pi / 2, args=(self.eccentricity,))
        return 4 * self.major_radius * integral

    @property
    def area(self) -> Quantity:
        """Area of the ellipse."""
        return np.pi * self.major_radius * self.minor_radius

    def transformed_copy(self, matrix: Matrix44) -> "Ellipse":
        """Create a transformed copy of the ellipse from a transformation matrix.

        Parameters
        ----------
        matrix : Matrix44
            4x4 transformation matrix to apply to the ellipse.

        Returns
        -------
        Ellipse
            New ellipse that is the transformed copy of the original ellipse.
        """
        new_point = self.origin.transform(matrix)
        new_reference = self._reference.transform(matrix)
        new_axis = self._axis.transform(matrix)
        return Ellipse(
            new_point,
            self.major_radius,
            self.minor_radius,
            UnitVector3D(new_reference[0:3]),
            UnitVector3D(new_axis[0:3]),
        )

    def parameterization(self) -> Parameterization:
        """Get the parametrization of the ellipse.

        The parameter of an ellipse specifies the clockwise angle around the axis
        (right-hand corkscrew law), with a zero parameter at ``dir_x`` and a period of 2*pi.

        Returns
        -------
        Parameterization
            Information about how the ellipse is parameterized.
        """
        return Parameterization(ParamForm.PERIODIC, ParamType.OTHER, Interval(0, 2 * np.pi))

    def contains_param(self, param: Real) -> bool:  # noqa: D102
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:  # noqa: D102
        raise NotImplementedError("contains_point() is not implemented.")


class EllipseEvaluation(CurveEvaluation):
    """Evaluate an ellipse at a given parameter.

    Parameters
    ----------
    ellipse: ~ansys.geometry.core.shapes.curves.ellipse.Ellipse
        Ellipse to evaluate.
    parameter: float, int
        Parameter to evaluate the ellipse at.
    """

    def __init__(self, ellipse: Ellipse, parameter: Real) -> None:
        """``Initialize the ``EllipseEvaluation`` class."""
        self._ellipse = ellipse
        self._parameter = parameter

    @property
    def ellipse(self) -> Ellipse:
        """Ellipse being evaluated."""
        return self._ellipse

    @property
    def parameter(self) -> Real:
        """Parameter that the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """Position of the evaluation.

        Returns
        -------
        Point3D
            Point that lies on the ellipse at this evaluation.
        """
        return (
            self.ellipse.origin
            + ((self.ellipse.major_radius * np.cos(self.parameter)) * self.ellipse.dir_x).m
            + ((self.ellipse.minor_radius * np.sin(self.parameter)) * self.ellipse.dir_y).m
        )

    @cached_property
    def tangent(self) -> UnitVector3D:
        """Tangent of the evaluation.

        Returns
        -------
        UnitVector3D
            Tangent unit vector to the ellipse at this evaluation.
        """
        return (
            (self.ellipse.minor_radius * np.cos(self.parameter) * self.ellipse.dir_y).m
            - (self.ellipse.major_radius * np.sin(self.parameter) * self.ellipse.dir_x).m
        ).normalize()

    @cached_property
    def normal(self) -> UnitVector3D:
        """Normal of the evaluation.

        Returns
        -------
        UnitVector3D
            Normal unit vector to the ellipse at this evaluation.
        """
        return UnitVector3D.from_points(self.ellipse.origin, self.position)

    @cached_property
    def first_derivative(self) -> Vector3D:
        """Girst derivative of the evaluation.

        The first derivative is in the direction of the tangent and has a magnitude
        equal to the velocity (rate of change of position) at that point.

        Returns
        -------
        Vector3D
            First derivative of the evaluation.
        """
        return (self.ellipse.minor_radius * np.cos(self.parameter) * self.ellipse.dir_y).m - (
            self.ellipse.major_radius * np.sin(self.parameter) * self.ellipse.dir_x
        ).m

    @cached_property
    def second_derivative(self) -> Vector3D:
        """Second derivative of the evaluation.

        Returns
        -------
        Vector3D
            Second derivative of the evaluation.
        """
        return (
            -self.ellipse.major_radius * np.cos(self.parameter) * self.ellipse.dir_x
            - self.ellipse.minor_radius * np.sin(self.parameter) * self.ellipse.dir_y
        ).m

    @cached_property
    def curvature(self) -> Real:
        """Curvature of the ellipse.

        Returns
        -------
        Real
            Curvature of the ellipse.
        """
        return self.second_derivative.magnitude / np.power(self.first_derivative.magnitude, 2)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the ``ProductInstance`` class."""

import logging
import os
from pathlib import Path
import signal
import socket

# Subprocess is needed to start the backend. But
# the input is controlled by the library. Excluding bandit check.
import subprocess  # nosec B404
from typing import TYPE_CHECKING

from ansys.geometry.core.connection.backend import ApiVersions, BackendType
from ansys.geometry.core.logger import LOG
from ansys.tools.path import get_available_ansys_installations, get_latest_ansys_installation

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.modeler import Modeler


WINDOWS_GEOMETRY_SERVICE_FOLDER = "GeometryService"
"""Default Geometry Service's folder name into the unified installer (DMS)."""

CORE_GEOMETRY_SERVICE_FOLDER = "GeometryService"
"""Default Geometry Service's folder name into the unified installer (Core Service)."""

DISCOVERY_FOLDER = "Discovery"
"""Default Discovery's folder name into the unified installer."""

SPACECLAIM_FOLDER = "scdm"
"""Default SpaceClaim's folder name into the unified installer."""

ADDINS_SUBFOLDER = "Addins"
"""Default global Addins's folder name into the unified installer."""

BACKEND_SUBFOLDER = "ApiServer"
"""Default backend's folder name into the ``ADDINS_SUBFOLDER`` folder."""

MANIFEST_FILENAME = "Presentation.ApiServerAddIn.Manifest.xml"
"""Default backend's add-in filename.

To be used only for local start of Ansys Discovery or Ansys SpaceClaim.
"""

GEOMETRY_SERVICE_EXE = "Presentation.ApiServerDMS.exe"
"""The Windows Geometry Service's filename (DMS)."""

CORE_GEOMETRY_SERVICE_EXE = "Presentation.ApiServerCoreService.exe"
"""The Windows Geometry Service's filename (Core Service)."""

DISCOVERY_EXE = "Discovery.exe"
"""The Ansys Discovery's filename."""

SPACECLAIM_EXE = "SpaceClaim.exe"
"""The Ansys SpaceClaim's filename."""

BACKEND_LOG_LEVEL_VARIABLE = "LOG_LEVEL"
"""The backend's log level environment variable for local start."""

BACKEND_TRACE_VARIABLE = "ENABLE_TRACE"
"""The backend's enable trace environment variable for local start."""

BACKEND_HOST_VARIABLE = "API_ADDRESS"
"""The backend's ip address environment variable for local start."""

BACKEND_PORT_VARIABLE = "API_PORT"
"""The backend's port number environment variable for local start."""

BACKEND_LOGS_FOLDER_VARIABLE = "ANS_DSCO_REMOTE_LOGS_FOLDER"
"""The backend's logs folder path to be used.

Only applicable to the Ansys Geometry Service.
"""

BACKEND_API_VERSION_VARIABLE = "API_VERSION"
"""The backend's api version environment variable for local start.

To be used only with Ansys Discovery and Ansys SpaceClaim.
"""

BACKEND_SPACECLAIM_OPTIONS = "--spaceclaim-options"
"""The additional argument for local Ansys Discovery start.

To be used only with Ansys Discovery.
"""

BACKEND_ADDIN_MANIFEST_ARGUMENT = "/ADDINMANIFESTFILE="
"""The argument to specify the backend's add-in manifest file's path.

To be used only with Ansys Discovery and Ansys SpaceClaim.
"""

BACKEND_SPACECLAIM_HIDDEN = "/Headless=True"
"""The argument to hide SpaceClaim's UI on the backend.

To be used only with Ansys SpaceClaim.
"""

BACKEND_SPACECLAIM_HIDDEN_ENVVAR_KEY = "SPACECLAIM_MODE"
"""SpaceClaim hidden backend's environment variable key.

To be used only with Ansys SpaceClaim.
"""

BACKEND_SPACECLAIM_HIDDEN_ENVVAR_VALUE = "2"
"""SpaceClaim hidden backend's environment variable value.

To be used only with Ansys SpaceClaim.
"""

BACKEND_DISCOVERY_HIDDEN = "--hidden"
"""The argument to hide Discovery's UI on the backend.

To be used only with Ansys Discovery.
"""

BACKEND_SPLASH_OFF = "/Splash=False"
"""The argument to specify the backend's add-in manifest file's path.

To be used only with Ansys Discovery and Ansys SpaceClaim.
"""

ANSYS_GEOMETRY_SERVICE_ROOT = "ANSYS_GEOMETRY_SERVICE_ROOT"
"""Local Geometry Service install location. This is for GeometryService and CoreGeometryService."""


class ProductInstance:
    """``ProductInstance`` class.

    This class is used as a handle for a local session of Ansys Product's backend: Discovery,
    Windows Geometry Service or SpaceClaim.

    Parameters
    ----------
    pid : int
        The local instance's process identifier. This allows to keep track of the process
        and close it if need be.
    """

    def __init__(self, pid: int):
        """Initialize ``ProductInstance`` class."""
        self._pid = pid

    def close(self) -> bool:
        """Close the process associated to the pid."""
        try:
            os.kill(self._pid, signal.SIGTERM)
        except OSError as oserr:
            LOG.error(str(oserr))
            return False

        return True


def prepare_and_start_backend(
    backend_type: BackendType,
    version: str | int = None,
    host: str = "localhost",
    port: int = None,
    enable_trace: bool = False,
    api_version: ApiVersions = ApiVersions.LATEST,
    timeout: int = 150,
    manifest_path: str = None,
    hidden: bool = False,
    server_log_level: int = 2,
    client_log_level: int = logging.INFO,
    server_logs_folder: str = None,
    client_log_file: str = None,
    specific_minimum_version: int = None,
    product_version: int | None = None,  # Deprecated, use `version` instead.
) -> "Modeler":
    """Start the requested service locally using the ``ProductInstance`` class.

    When calling this method, a standalone service or product session is started.
    By default, if an endpoint is specified (by defining `host` and `port` parameters)
    but the endpoint is not available, the startup will fail. Otherwise, it will try to
    launch its own service.

    Parameters
    ----------
    version : str | int, optional
        The product version to be started. Goes from v24.1 to
        the latest. Default is ``None``.
        If a specific product version is requested but not installed locally,
        a SystemError will be raised.
    host: str, optional
        IP address at which the Geometry service will be deployed. By default,
        its value will be ``localhost``.
    port : int, optional
        Port at which the Geometry service will be deployed. By default, its
        value will be ``None``.
    enable_trace : bool, optional
        Boolean enabling the logs trace on the Geometry service console window.
        By default its value is ``False``.
    api_version: ``ApiVersions``, optional
        The backend's API version to be used at runtime. Goes from API v21 to
        the latest. Default is ``ApiVersions.LATEST``.
    timeout : int, optional
        Timeout for starting the backend startup process. The default is 150.
    manifest_path : str, optional
        Used to specify a manifest file path for the ApiServerAddin. This way,
        it is possible to run an ApiServerAddin from a version an older product
        version. Only applicable for Ansys Discovery and Ansys SpaceClaim.
    hidden : starts the product hiding its UI. Default is ``False``.
    server_log_level : int, optional
        Backend's log level from 0 to 3:
            0: Chatterbox
            1: Debug
            2: Warning
            3: Error

        The default is ``2`` (Warning).
    client_log_level : int, optional
        Logging level to apply to the client. By default, INFO level is used.
        Use the logging module's levels: DEBUG, INFO, WARNING, ERROR, CRITICAL.
    server_logs_folder : str, optional
        Sets the backend's logs folder path. If nothing is defined,
        the backend will use its default path.
    client_log_file : str, optional
        Sets the client's log file path. If nothing is defined,
        the client will log to the console.
    specific_minimum_version : int, optional
        Sets a specific minimum version to be checked. If this is not defined,
        the minimum version will be set to 24.1.0.
    product_version: ``int``, optional
        The product version to be started. Deprecated, use `version` instead.

    Returns
    -------
    Modeler
        Instance of the Geometry service.

    Raises
    ------
    ConnectionError
        If the specified endpoint is already in use, a connection error will be raised.
    SystemError
        If there is not an Ansys product 24.1 version or later installed
        or if a specific product's version is requested but not installed locally then
        a SystemError will be raised.
    """
    from ansys.geometry.core.modeler import Modeler

    # Handle deprecated product_version argument
    if product_version is not None and version is not None:
        raise ValueError(
            "Both 'product_version' and 'version' arguments are provided. "
            "Please use only 'version'."
        )
    if product_version is not None:
        LOG.warning("The 'product_version' argument is deprecated. Please use 'version' instead.")
        version = product_version

    if os.name != "nt" and not BackendType.is_linux_service(backend_type):  # pragma: no cover
        raise RuntimeError(
            "Method 'prepare_and_start_backend' is only available on Windows."
            "A Linux version is only available for the Core Geometry Service."
        )

    port = _check_port_or_get_one(port)
    installations = get_available_ansys_installations()
    if os.getenv(ANSYS_GEOMETRY_SERVICE_ROOT) is not None and backend_type in (
        BackendType.WINDOWS_SERVICE,
        BackendType.LINUX_SERVICE,
        BackendType.CORE_WINDOWS,
        BackendType.CORE_LINUX,
    ):
        # If the user has set the ANSYS_GEOMETRY_SERVICE_ROOT environment variable,
        # we will use it as the root folder for the Geometry Service.
        pass
    else:
        if version is not None:
            # Sanitize the version input to ensure it's an integer.
            try:
                version = int(version)
            except ValueError:
                raise ValueError(
                    "The 'version' argument must be an integer representing the product version."
                )

            try:
                _check_version_is_available(version, installations)
            except SystemError as serr:
                # The user requested a version as a Student version...
                # Let's negate it and try again... if this works, we override the
                # version variable.
                try:
                    _check_version_is_available(-version, installations)
                except SystemError:
                    # The student version is not installed either... raise the original error.
                    raise serr

                version = -version
        else:
            version = get_latest_ansys_installation()[0]

        # Verify that the minimum version is installed.
        _check_minimal_versions(version, specific_minimum_version)

    # If Windows Service is requested, BUT version is not 2025R1 or earlier, we will have
    # to change the backend type to CORE_WINDOWS and throw a warning.
    if backend_type == BackendType.WINDOWS_SERVICE and version > 251:
        LOG.warning(
            "DMS Windows Service is not available for Ansys versions 2025R2 and later. "
            "Switching to Core Windows Service."
        )
        backend_type = BackendType.CORE_WINDOWS

    if server_logs_folder is not None:
        # Verify that the user has write permissions to the folder and that it exists.
        try:
            # Make sure the folder exists...
            Path(server_logs_folder).mkdir(parents=True, exist_ok=True)
            # Create a file to test write permissions...
            Path(server_logs_folder, ".verify").touch(exist_ok=True)
        except PermissionError:
            raise RuntimeError(
                "User does not have write permissions to the logs folder "
                f"{Path(server_logs_folder).resolve()}"
            )

    args = []
    env_copy = _get_common_env(
        host=host,
        port=port,
        enable_trace=enable_trace,
        server_log_level=server_log_level,
        server_logs_folder=server_logs_folder,
    )

    if backend_type == BackendType.DISCOVERY:
        args.append(Path(installations[version], DISCOVERY_FOLDER, DISCOVERY_EXE))
        if hidden is True:
            args.append(BACKEND_DISCOVERY_HIDDEN)

        # Here begins the spaceclaim arguments.
        args.append(BACKEND_SPACECLAIM_OPTIONS)
        args.append(
            BACKEND_ADDIN_MANIFEST_ARGUMENT
            + _manifest_path_provider(version, installations, manifest_path)
        )
        env_copy[BACKEND_API_VERSION_VARIABLE] = str(api_version)

    elif backend_type == BackendType.SPACECLAIM:
        args.append(Path(installations[version], SPACECLAIM_FOLDER, SPACECLAIM_EXE))
        if hidden is True:
            args.append(BACKEND_SPACECLAIM_HIDDEN)
            args.append(BACKEND_SPLASH_OFF)
        args.append(
            BACKEND_ADDIN_MANIFEST_ARGUMENT
            + _manifest_path_provider(version, installations, manifest_path)
        )
        env_copy[BACKEND_API_VERSION_VARIABLE] = str(api_version)
        if BACKEND_SPACECLAIM_HIDDEN_ENVVAR_KEY not in env_copy:
            env_copy[BACKEND_SPACECLAIM_HIDDEN_ENVVAR_KEY] = BACKEND_SPACECLAIM_HIDDEN_ENVVAR_VALUE
        else:
            LOG.warning(
                f"Environment variable '{BACKEND_SPACECLAIM_HIDDEN_ENVVAR_KEY}' already exists. "
                f"Using value '{env_copy[BACKEND_SPACECLAIM_HIDDEN_ENVVAR_KEY]}'."
            )

    elif backend_type == BackendType.WINDOWS_SERVICE:
        root_service_folder = os.getenv(ANSYS_GEOMETRY_SERVICE_ROOT)
        if root_service_folder is None:
            root_service_folder = Path(installations[version], WINDOWS_GEOMETRY_SERVICE_FOLDER)
        else:
            root_service_folder = Path(root_service_folder)
        args.append(
            Path(
                root_service_folder,
                GEOMETRY_SERVICE_EXE,
            )
        )
    # This should be modified to Windows Core Service in the future
    elif BackendType.is_core_service(backend_type):
        # Define several Ansys Geometry Core Service folders needed
        root_service_folder = os.getenv(ANSYS_GEOMETRY_SERVICE_ROOT)
        if root_service_folder is None:
            root_service_folder = Path(installations[version], CORE_GEOMETRY_SERVICE_FOLDER)
        else:
            root_service_folder = Path(root_service_folder)
        native_folder = root_service_folder / "Native"
        cad_integration_folder = root_service_folder / "CADIntegration"
        cad_integration_folder_bin = cad_integration_folder / "bin"
        schema_folder = root_service_folder / "Schema"

        # Adapt the native folder to the OS
        # The native folder is different for Windows and Linux.
        if os.name == "nt":
            native_folder = native_folder / "Windows"
        else:
            native_folder = native_folder / "Linux"

        # Set the environment variables for the Ansys Geometry Core Service launch
        # ANS_DSCO_REMOTE_IP should be variable "host" directly, but not working...
        env_copy["ANS_DSCO_REMOTE_IP"] = "127.0.0.1" if host == "localhost" else host
        env_copy["ANS_DSCO_REMOTE_PORT"] = str(port)
        env_copy["P_SCHEMA"] = schema_folder.as_posix()
        env_copy["ANSYS_CI_INSTALL"] = cad_integration_folder.as_posix()
        if version:
            env_copy[f"ANSYSCL{version}_DIR"] = (root_service_folder / "licensingclient").as_posix()
        else:
            # Env var is passed... but no product version is defined. We define all of them.
            # Starting on 251... We can use the values in the enum ApiVersions.
            for api_version in ApiVersions:
                if api_version.value < 251:
                    continue
                env_copy[f"ANSYSCL{api_version.value}_DIR"] = (
                    root_service_folder / "licensingclient"
                ).as_posix()

        # License server - for Core Geometry Service. If not defined, it will use the default one.
        if "LICENSE_SERVER" in os.environ:
            pass  # Do nothing... the user has defined the license server.
        elif "ANSRV_GEO_LICENSE_SERVER" in os.environ:
            env_copy["LICENSE_SERVER"] = os.getenv("ANSRV_GEO_LICENSE_SERVER")
        else:
            env_copy["LICENSE_SERVER"] = os.getenv("ANSYSLMD_LICENSE_FILE", "1055@localhost")

        # Adapt the path environment variable to the OS and
        # modify the PATH/LD_LIBRARY_PATH variable to include the path
        # to the Ansys Geometry Core Service
        path_env_var = "PATH" if os.name == "nt" else "LD_LIBRARY_PATH"
        env_copy[path_env_var] = os.pathsep.join(
            [
                root_service_folder.as_posix(),
                native_folder.as_posix(),
                cad_integration_folder_bin.as_posix(),
                env_copy.get(path_env_var, ""),
            ]
        )

        if os.name == "nt":
            # For Windows, we need to use the exe file to launch the Core Geometry Service
            args.append(
                Path(
                    root_service_folder,
                    CORE_GEOMETRY_SERVICE_EXE,
                )
            )
        else:
            # Verify dotnet is installed
            import shutil

            if not shutil.which("dotnet"):
                raise RuntimeError(
                    "Cannot find 'dotnet' command. "
                    "Please install 'dotnet' to use the Ansys Geometry Core Service. "
                    "At least dotnet 8.0 is required."
                )

            # At least dotnet 8.0 is required
            # private method and controlled input by library - excluding bandit check.
            if subprocess.check_output(["dotnet", "--version"]).decode("utf-8").split(".")[0] < "8":  # nosec B607, B603
                raise RuntimeError(
                    "Ansys Geometry Core Service requires at least dotnet 8.0. "
                    "Please install a compatible version."
                )

            # For Linux, we need to use the dotnet command to launch the Core Geometry Service
            args.append("dotnet")
            args.append(
                Path(
                    root_service_folder,
                    CORE_GEOMETRY_SERVICE_EXE.replace(".exe", ".dll"),
                )
            )
    else:
        raise RuntimeError(
            f"Cannot connect to backend {backend_type.name} using ``prepare_and_start_backend()``"
        )

    LOG.info(f"Launching ProductInstance for {backend_type.name}")
    LOG.debug(f"Args: {args}")
    LOG.debug(f"Environment variables: {env_copy}")

    instance = ProductInstance(__start_program(args, env_copy).pid)

    # Verify that the backend is ready to accept connections
    # before returning the Modeler instance.
    LOG.info("Waiting for backend to be ready...")
    _wait_for_backend(host, port, timeout)

    return Modeler(
        host=host,
        port=port,
        timeout=timeout,
        product_instance=instance,
        logging_level=client_log_level,
        logging_file=client_log_file,
    )


def get_available_port() -> int:
    """Return an available port to be used.

    Returns
    -------
    int
        The available port.
    """
    sock = socket.socket()
    sock.bind((socket.gethostname(), 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def _wait_for_backend(host: str, port: int, timeout: int):
    """Check if the backend is ready to accept connections.

    Parameters
    ----------
    host : str
        The backend's ip address.
    port : int
        The backend's port number.
    timeout : int
        The timeout in seconds.
    """
    import time

    start_time = time.time()
    while time.time() - start_time < timeout:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex((host, port)) == 0:
                LOG.debug("Backend is ready to accept connections.")
                return
            else:
                LOG.debug("Still waiting for backend to be ready... Retrying in 5 seconds.")
                time.sleep(5)

    raise ConnectionError("Timeout while waiting for backend to be ready.")


def _is_port_available(port: int, host: str = "localhost") -> bool:
    """Check whether the argument port is available.

    The optional argument is the ip address where to check port availability.
    Its default is ``localhost``.
    """
    if port == 0:
        return False  # Port 0 is reserved and cannot be used directly.
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            sock.bind((host, port))
            return True
        except socket.error:
            return False


def _manifest_path_provider(
    version: int, available_installations: dict[str, str], manifest_path: str = None
) -> str:
    """Return the ApiServer's add-in manifest file path."""
    if manifest_path:
        if Path(manifest_path).exists():
            return manifest_path
        else:
            LOG.warning(
                "Specified manifest file's path does not exist. Taking install default path."
            )

    # Default manifest path
    def_manifest_path = Path(
        available_installations[version], ADDINS_SUBFOLDER, BACKEND_SUBFOLDER, MANIFEST_FILENAME
    )

    if def_manifest_path.exists():
        return def_manifest_path.as_posix()
    else:
        msg = (
            "Default manifest file's path does not exist."
            " Please specify a valid manifest."
            f" The ApiServer Add-In seems to be missing in {def_manifest_path.parent}"
        )
        LOG.error(msg)
        raise RuntimeError(msg)


def __start_program(args: list[str], local_env: dict[str, str]) -> subprocess.Popen:
    """Start the program.

    Parameters
    ----------
    args : list[str]
        List of arguments to be passed to the program. The first list's item shall
        be the program path.
    local_env : dict[str,str]
        Environment variables to be passed to the program.

    Returns
    -------
    subprocess.Popen
        The subprocess object.

    Notes
    -----
    The path is the first item of the ``args`` array argument.
    """
    # private method and controlled input by library - excluding bandit check.
    return subprocess.Popen(  # nosec B603
        args,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        env=local_env,
    )


def _check_minimal_versions(
    latest_installed_version: int, specific_minimum_version: int | None
) -> None:
    """Check client is compatible with Ansys Products.

    Check that at least V241 is installed. Or, if a specific version is requested,
    check that it is installed.
    """
    min_ver = specific_minimum_version or 241
    if abs(latest_installed_version) < min_ver:
        # Split the version into its components.
        major, minor = divmod(min_ver, 10)
        msg = (
            f"PyAnsys Geometry is compatible with Ansys Products from version {major}.{minor}.0. "
            + f"Please install Ansys products {major}.{minor}.0 or later."
        )
        raise SystemError(msg)


def _check_version_is_available(version: int, installations: dict[int, str]) -> None:
    """Check that the requested version for launcher is installed."""
    if version not in installations:
        msg = (
            f"The requested Ansys product's version {version} is not available, "
            + "please specify a different version."
        )
        raise SystemError(msg)


def _check_port_or_get_one(port: int) -> int:
    """If a ``port`` argument is specified, check that it's free.

    If not, raise an error.

    If ``port`` is None, return an available port by calling ``get_available_port``.
    """
    if port:
        if _is_port_available(port):
            return port
        else:
            msg = f"Port {port} is already in use. Please specify a different one."
            raise ConnectionError(msg)
    else:
        return get_available_port()


def _get_common_env(
    host: str,
    port: int,
    enable_trace: bool,
    server_log_level: int,
    server_logs_folder: str = None,
) -> dict[str, str]:
    """Make a copy of the actual system's environment.

    Then update or create some environment variables with the provided
    arguments.
    """
    env_copy = os.environ.copy()

    env_copy[BACKEND_HOST_VARIABLE] = host
    env_copy[BACKEND_PORT_VARIABLE] = f"{port}"
    env_copy[BACKEND_LOG_LEVEL_VARIABLE] = f"{server_log_level}"

    if server_logs_folder is not None:
        env_copy[BACKEND_LOGS_FOLDER_VARIABLE] = f"{server_logs_folder}"

    if enable_trace:
        env_copy[BACKEND_TRACE_VARIABLE] = "1"  # for backward compatibility
        if server_log_level > 0:
            LOG.warning("Enabling trace mode will override the log level to 0 (Chatterbox).")
            env_copy[BACKEND_LOG_LEVEL_VARIABLE] = "0"

    return env_copy



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ``MaterialProperty`` class."""

from enum import Enum, unique

from beartype import beartype as check_input_types
from pint import Quantity

from ansys.geometry.core.typing import Real


@unique
class MaterialPropertyType(Enum):
    """Enum holding the possible values for ``MaterialProperty`` objects."""

    DENSITY = "Density"
    ELASTIC_MODULUS = "ElasticModulus"
    POISSON_RATIO = "PoissonsRatio"
    SHEAR_MODULUS = "ShearModulus"
    SPECIFIC_HEAT = "SpecificHeat"
    TENSILE_STRENGTH = "TensileStrength"
    THERMAL_CONDUCTIVITY = "ThermalConductivity"

    @staticmethod
    def from_id(id: str) -> "MaterialPropertyType":
        """Return the ``MaterialPropertyType`` value from the service.

        Parameters
        ----------
        id : str
            Geometry Service string representation of a property type.

        Returns
        -------
        MaterialPropertyType
            Common name for property type.
        """
        relations = {
            "General.Density.Mass": "Density",
            "Linear.Isotropic.Emodulus": "ElasticModulus",
            "Linear.Isotropic.Poisson": "PoissonsRatio",
            "Linear.Isotropic.Gmodulus": "ShearModulus",
            "Heat.Iso.Cp": "SpecificHeat",
            "Nonlinear.vonMises.Ultimate": "TensileStrength",
            "Heat.Iso.Conduction": "ThermalConductivity",
        }
        return MaterialPropertyType(relations.get(id))


class MaterialProperty:
    """Provides the data structure for a material property.

    Parameters
    ----------
    type : MaterialPropertyType | str
        Type of the material property. If the type is a string, it must be a valid
        material property type - though it might not be supported by the MaterialPropertyType
        enum.
    name: str
        Material property name.
    quantity: ~pint.Quantity | Real
        Value and unit in case of a supported Quantity. If the type is not supported, it
        must be a Real value (float or integer).
    """

    @check_input_types
    def __init__(
        self,
        type: MaterialPropertyType | str,
        name: str,
        quantity: Quantity | Real,
    ):
        """Initialize ``MaterialProperty`` class."""
        self._type = type
        self._name = name
        self._quantity = quantity

    @property
    def type(self) -> MaterialPropertyType | str:
        """Material property ID.

        If the type is not supported, it will be a string.
        """
        return self._type

    @property
    def name(self) -> str:
        """Material property name."""
        return self._name

    @property
    def quantity(self) -> Quantity | Real:
        """Material property quantity and unit.

        If the type is not supported, it will be a Real value (float or
        integer).
        """
        return self._quantity



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a nurbs sketch curve."""

from typing import TYPE_CHECKING

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.checks import graphics_required
from ansys.geometry.core.sketch.edge import SketchEdge
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    import geomdl.NURBS as geomdl_nurbs  # noqa: N811
    import pyvista as pv


class SketchNurbs(SketchEdge):
    """Represents a NURBS sketch curve.

    Warnings
    --------
    NURBS sketching is only supported in 26R1 and later versions of Ansys.

    Notes
    -----
    This class is a wrapper around the NURBS curve class from the `geomdl` library.
    By leveraging the `geomdl` library, this class provides a high-level interface
    to create and manipulate NURBS curves. The `geomdl` library is a powerful
    library for working with NURBS curves and surfaces. For more information, see
    https://pypi.org/project/geomdl/.
    """

    def __init__(self):
        """Initialize the NURBS sketch curve."""
        super().__init__()
        try:
            import geomdl.NURBS as geomdl_nurbs  # noqa: N811
        except ImportError as e:  # pragma: no cover
            raise ImportError(
                "The `geomdl` library is required to use the NURBSCurve class. "
                "Please install it using `pip install geomdl`."
            ) from e

        self._nurbs_curve = geomdl_nurbs.Curve()

    @property
    def geomdl_nurbs_curve(self) -> "geomdl_nurbs.Curve":
        """Get the underlying NURBS curve.

        Notes
        -----
        This property gives access to the full functionality of the NURBS curve
        coming from the `geomdl` library. Use with caution.
        """
        return self._nurbs_curve

    @property
    def control_points(self) -> list[Point2D]:
        """Get the control points of the curve."""
        return [Point2D(point) for point in self._nurbs_curve.ctrlpts]

    @property
    def degree(self) -> int:
        """Get the degree of the curve."""
        return self._nurbs_curve.degree

    @property
    def knots(self) -> list[Real]:
        """Get the knot vector of the curve."""
        return self._nurbs_curve.knotvector

    @property
    def weights(self) -> list[Real]:
        """Get the weights of the control points."""
        return self._nurbs_curve.weights

    @property
    def start(self) -> Point2D:
        """Get the start point of the curve."""
        return Point2D(self._nurbs_curve.evaluate_single(0.0))

    @property
    def end(self) -> Point2D:
        """Get the end point of the curve."""
        return Point2D(self._nurbs_curve.evaluate_single(1.0))

    @property
    @graphics_required
    def visualization_polydata(self) -> "pv.PolyData":
        """Get the VTK polydata representation for PyVista visualization.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.

        Notes
        -----
        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.
        """
        import pyvista as pv

        # Sample points along the curve
        params = np.linspace(0, 1, 100)
        points = [self._nurbs_curve.evaluate_single(u) for u in params]  # For 2D: [x, y]

        # Add a zero z-coordinate for PyVista (only supports 3D points)
        points = [(*pt, 0.0) for pt in points]

        # Create PolyData and add the line
        polydata = pv.PolyData(points)
        polydata.lines = [len(points)] + list(range(len(points)))

        return polydata

    def contains_point(self, point: Point2D, tolerance: Real = 1e-6) -> bool:
        """Check if the curve contains a given point within a specified tolerance.

        Parameters
        ----------
        point : Point2D
            The point to check.
        tolerance : Real, optional
            The tolerance for the containment check, by default 1e-6.

        Returns
        -------
        bool
            True if the curve contains the point within the tolerance, False otherwise.
        """
        # Sample points along the curve
        params = np.linspace(0, 1, 200)
        sampled = [self._nurbs_curve.evaluate_single(u) for u in params]

        # Check if any sampled point is close to the target point
        return any(np.linalg.norm(np.array(pt) - np.array(point)) < tolerance for pt in sampled)

    @classmethod
    @check_input_types
    def fit_curve_from_points(
        cls,
        points: list[Point2D],
        degree: int = 3,
    ) -> "SketchNurbs":
        """Fit a NURBS curve to a set of points.

        Parameters
        ----------
        points : list[Point2D]
            The points to fit the curve to.
        degree : int, optional
            The degree of the NURBS curve, by default 3.

        Returns
        -------
        SketchNurbs
            A new instance of SketchNurbs fitted to the given points.
        """
        from geomdl import fitting

        # Check degree compared to number of points provided
        if degree < 1:
            raise ValueError("Degree must be at least 1.")
        if len(points) == 2:
            degree = 1  # Force linear interpolation for two points
        if len(points) == 3:
            degree = 2  # Force quadratic interpolation for three points
        if degree >= len(points):
            raise ValueError(
                f"Degree {degree} is too high for the number of points provided: {len(points)}."
            )

        curve = fitting.interpolate_curve(
            [[*pt] for pt in points],  # Convert Point2D to list of coordinates
            degree=degree,
        )

        # Construct the NURBSCurve object
        nurbs_curve = cls()
        nurbs_curve._nurbs_curve.degree = curve.degree
        nurbs_curve._nurbs_curve.ctrlpts = [Point2D(entry) for entry in curve.ctrlpts]
        nurbs_curve._nurbs_curve.knotvector = curve.knotvector
        nurbs_curve._nurbs_curve.weights = curve.weights

        # Verify the curve is valid
        try:
            nurbs_curve._nurbs_curve._check_variables()
        except ValueError as e:
            raise ValueError(f"Invalid NURBS curve: {e}")

        return nurbs_curve



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the Driving Dimension service implementation for v0."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.driving_dimensions import GRPCDrivingDimensionsService
from .conversions import (
    from_driving_dimension_to_grpc_driving_dimension,
    from_grpc_driving_dimension_to_driving_dimension,
    from_grpc_update_status_to_parameter_update_status,
)


class GRPCDrivingDimensionsServiceV0(GRPCDrivingDimensionsService):
    """Driving Dimension service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    driving dimension service. It is specifically designed for the v0 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.dbu.v0.drivingdimensions_pb2_grpc import DrivingDimensionsStub

        self.stub = DrivingDimensionsStub(channel)

    @protect_grpc
    def get_all_parameters(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.drivingdimensions_pb2 import GetAllRequest

        # Call the gRPC service
        response = self.stub.GetAll(GetAllRequest())

        # Return the response - formatted as a dictionary
        return {
            "parameters": [
                from_grpc_driving_dimension_to_driving_dimension(param)
                for param in response.driving_dimensions
            ],
        }

    @protect_grpc
    def set_parameter(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.dbu.v0.drivingdimensions_pb2 import UpdateRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = UpdateRequest(
            driving_dimension=from_driving_dimension_to_grpc_driving_dimension(
                kwargs["driving_dimension"]
            ),
        )

        # Call the gRPC service
        response = self.stub.UpdateParameter(request)

        # Return the response - formatted as a dictionary
        return {
            "status": from_grpc_update_status_to_parameter_update_status(response.status),
        }



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing a bounding box."""

import sys
from typing import Union

from beartype import beartype as check_input_types

from ansys.geometry.core.math.misc import intersect_interval
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.misc.accuracy import Accuracy
from ansys.geometry.core.misc.checks import deprecated_method
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS
from ansys.geometry.core.typing import Real


class BoundingBox2D:
    """Maintains the X and Y dimensions.

    Parameters
    ----------
    x_min : Real
        Minimum value for the x-dimensional bounds.
    x_max : Real
        Maximum value for the x-dimensional bounds.
    y_min : Real
        Minimum value for the y-dimensional bounds.
    y_max : Real
        Maximum value for the y-dimensional bounds.
    """

    @check_input_types
    @deprecated_method(alternative="BoundingBox", version="0.10.0", remove="0.12.0")
    def __init__(
        self,
        x_min: Real = sys.float_info.max,
        x_max: Real = sys.float_info.min,
        y_min: Real = sys.float_info.max,
        y_max: Real = sys.float_info.min,
    ):
        """Initialize the ``BoundingBox2D`` class."""
        self._x_min = x_min
        self._x_max = x_max
        self._y_min = y_min
        self._y_max = y_max

    @property
    def x_min(self) -> Real:
        """Minimum value of X-dimensional bounds.

        Returns
        -------
        Real
            Minimum value of the X-dimensional bounds.
        """
        return self._x_min

    @property
    def x_max(self) -> Real:
        """Maximum value of the X-dimensional bounds.

        Returns
        -------
        Real
            Maximum value of the X-dimensional bounds.
        """
        return self._x_max

    @property
    def y_min(self) -> Real:
        """Minimum value of Y-dimensional bounds.

        Returns
        -------
        Real
            Minimum value of Y-dimensional bounds.
        """
        return self._y_min

    @property
    def y_max(self) -> Real:
        """Maximum value of Y-dimensional bounds.

        Returns
        -------
        Real
            Maximum value of Y-dimensional bounds.
        """
        return self._y_max

    @check_input_types
    def add_point(self, point: Point2D) -> None:
        """Extend the ranges of the bounding box to include a point.

        Parameters
        ----------
        point : Point2D
            Point to include within the bounds.

        Notes
        -----
        This method is only applicable if the point components are outside
        the current bounds.
        """
        self.add_point_components(
            point.x.m_as(DEFAULT_UNITS.LENGTH), point.y.m_as(DEFAULT_UNITS.LENGTH)
        )

    @check_input_types
    def add_point_components(self, x: Real, y: Real) -> None:
        """Extend the ranges of the bounding box to include the X and Y values.

        Parameters
        ----------
        x : Real
            Point X component to include within the bounds.
        y : Real
            Point Y component to include within the bounds.

        Notes
        -----
        This method is only applicable if the point components are outside
        the current bounds.
        """
        self._x_min = x if x < self._x_min else self._x_min
        self._x_max = x if x > self._x_max else self._x_max
        self._y_min = y if y < self._y_min else self._y_min
        self._y_max = y if y > self._y_max else self._y_max

    @check_input_types
    def add_points(self, points: list[Point2D]) -> None:
        """Extend the ranges of the bounding box to include given points.

        Parameters
        ----------
        points : list[Point2D]
            List of points to include within the bounds.
        """
        for point in points:
            self.add_point(point)

    @check_input_types
    def contains_point(self, point: Point2D) -> bool:
        """Evaluate whether a point lies within the X and Y range bounds.

        Parameters
        ----------
        point : Point2D
            Point to compare against the bounds.

        Returns
        -------
        bool
            ``True`` if the point is contained in the bounding box. Otherwise, ``False``.
        """
        return self.contains_point_components(
            point.x.m_as(DEFAULT_UNITS.LENGTH), point.y.m_as(DEFAULT_UNITS.LENGTH)
        )

    @check_input_types
    def contains_point_components(self, x: Real, y: Real) -> bool:
        """Check if point components are within the X and Y range bounds.

        Parameters
        ----------
        x : Real
            Point X component to compare against the bounds.
        y : Real
            Point Y component to compare against the bounds.

        Returns
        -------
        bool
            ``True`` if the components are contained in the bounding box. Otherwise, ``False``.
        """
        return (
            Accuracy.length_is_greater_than_or_equal(x, self._x_min)
            and Accuracy.length_is_greater_than_or_equal(y, self._y_min)
            and Accuracy.length_is_less_than_or_equal(x, self._x_max)
            and Accuracy.length_is_less_than_or_equal(y, self._y_max)
        )

    @check_input_types
    def __eq__(self, other: "BoundingBox2D") -> bool:
        """Equals operator for the ``BoundingBox2D`` class."""
        return (
            self.x_min == other.x_min
            and self.x_max == other.x_max
            and self.y_min == other.y_min
            and self.y_max == other.y_max
        )

    @check_input_types
    def __ne__(self, other: "BoundingBox2D") -> bool:
        """Not equals operator for the ``BoundingBox2D`` class."""
        return not self == other

    @staticmethod
    def intersect_bboxes(
        box_1: "BoundingBox2D", box_2: "BoundingBox2D"
    ) -> Union[None, "BoundingBox2D"]:
        """Find the intersection of 2 BoundingBox2D objects.

        Parameters
        ----------
        box_1: BoundingBox2D
            The box to consider the intersection of with respect to box_2.
        box_2: BoundingBox2D
            The box to consider the intersection of with respect to box_1.

        Returns
        -------
        BoundingBox2D:
            The box representing the intersection of the two passed in boxes.
        """
        intersect, min_x, max_x = intersect_interval(
            box_1.x_min, box_2.x_min, box_1.x_max, box_2.x_max
        )
        if not intersect:
            return None

        intersect, min_y, max_y = intersect_interval(
            box_1.y_min, box_2.y_min, box_1.y_max, box_2.y_max
        )
        if not intersect:
            return None

        return BoundingBox2D(min_x, max_x, min_y, max_y)


class BoundingBox:
    """Maintains the box structure for Bounding Boxes.

    Parameters
    ----------
    min_corner : Point3D
        Minimum corner for the box.
    max_corner : Point3D
        Maximum corner for the box.
    center : Point3D
        Center of the box.
    """

    @check_input_types
    def __init__(self, min_corner: Point3D, max_corner: Point3D, center: Point3D = None):
        """Initialize the ``BoundingBox`` class."""
        self._min_corner = min_corner
        self._max_corner = max_corner

        if center is not None:
            self._center = center
        else:
            mid_x = (max_corner.x.m - min_corner.x.m) / 2
            mid_y = (max_corner.y.m - min_corner.y.m) / 2
            mid_z = (max_corner.z.m - min_corner.z.m) / 2
            self._center = Point3D([mid_x, mid_y, mid_z])

    @property
    def min_corner(self) -> Point3D:
        """Minimum corner of the bounding box."""
        return self._min_corner

    @property
    def max_corner(self) -> Point3D:
        """Maximum corner of the bounding box."""
        return self._max_corner

    @property
    def center(self) -> Point3D:
        """Center of the bounding box."""
        return self._center

    @check_input_types
    def contains_point(self, point: Point3D) -> bool:
        """Evaluate whether a point lies within the box.

        Parameters
        ----------
        point : Point3D
            Point to compare against the bounds.

        Returns
        -------
        bool
            ``True`` if the point is contained in the bounding box. Otherwise, ``False``.
        """
        return self.contains_point_components(
            point.x.m_as(DEFAULT_UNITS.LENGTH),
            point.y.m_as(DEFAULT_UNITS.LENGTH),
            point.z.m_as(DEFAULT_UNITS.LENGTH),
        )

    @check_input_types
    def contains_point_components(self, x: Real, y: Real, z: Real) -> bool:
        """Check if point components are within box.

        Parameters
        ----------
        x : Real
            Point X component to compare against the bounds.
        y : Real
            Point Y component to compare against the bounds.
        z : Real
            Point Z component to compare against the bounds.

        Returns
        -------
        bool
            ``True`` if the components are contained in the bounding box. Otherwise, ``False``.
        """
        return (
            Accuracy.length_is_greater_than_or_equal(x, self._min_corner.x.m)
            and Accuracy.length_is_greater_than_or_equal(y, self._min_corner.y.m)
            and Accuracy.length_is_greater_than_or_equal(z, self._min_corner.z.m)
            and Accuracy.length_is_less_than_or_equal(x, self._max_corner.x.m)
            and Accuracy.length_is_less_than_or_equal(y, self._max_corner.y.m)
            and Accuracy.length_is_less_than_or_equal(z, self._max_corner.z.m)
        )

    @check_input_types
    def __eq__(self, other: "BoundingBox") -> bool:
        """Equals operator for the ``BoundingBox`` class."""
        return (
            self._min_corner.x == other._min_corner.x
            and self._max_corner.x == other._max_corner.x
            and self._min_corner.y == other._min_corner.y
            and self._max_corner.y == other._max_corner.y
            and self._min_corner.z == other._min_corner.z
            and self._max_corner.z == other._max_corner.z
        )

    @check_input_types
    def __ne__(self, other: "BoundingBox") -> bool:
        """Not equals operator for the ``BoundingBox`` class."""
        return not self == other

    @staticmethod
    def intersect_bboxes(box_1: "BoundingBox", box_2: "BoundingBox") -> Union[None, "BoundingBox"]:
        """Find the intersection of 2 BoundingBox objects.

        Parameters
        ----------
        box_1: BoundingBox
            The box to consider the intersection of with respect to box_2.
        box_2: BoundingBox
            The box to consider the intersection of with respect to box_1.

        Returns
        -------
        BoundingBox:
            The box representing the intersection of the two passed in boxes.
        """
        intersect, min_x, max_x = intersect_interval(
            box_1._min_corner.x.m,
            box_2._min_corner.x.m,
            box_1._max_corner.x.m,
            box_2._max_corner.x.m,
        )
        if not intersect:
            return None

        intersect, min_y, max_y = intersect_interval(
            box_1._min_corner.y.m,
            box_2._min_corner.y.m,
            box_1._max_corner.y.m,
            box_2._max_corner.y.m,
        )
        if not intersect:
            return None

        intersect, min_z, max_z = intersect_interval(
            box_1._min_corner.z.m,
            box_2._min_corner.z.m,
            box_1._max_corner.z.m,
            box_2._max_corner.z.m,
        )
        if not intersect:
            return None

        min_point = Point3D([min_x, min_y, min_z])
        max_point = Point3D([max_x, max_y, max_z])
        return BoundingBox(min_point, max_point)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the bodies service implementation for v1."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.bodies import GRPCBodyService


class GRPCBodyServiceV1(GRPCBodyService):  # pragma: no cover
    """Body service for gRPC communication with the Geometry server.

    This class provides methods to create and manipulate bodies in the
    Geometry server using gRPC. It is specifically designed for the v1
    version of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):
        """Initialize the BodyService with the gRPC stub."""
        from ansys.api.geometry.v1.bodies_pb2_grpc import BodiesStub

        self.stub = BodiesStub(channel)

    @protect_grpc
    def create_sphere_body(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_extruded_body(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_sweeping_profile_body(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_sweeping_chain(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def sweep_with_guide(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_extruded_body_from_face_profile(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_extruded_body_from_loft_profiles(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_planar_body(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_body_from_face(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_surface_body(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def create_surface_body_from_trimmed_curves(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def translate(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def delete(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def is_suppressed(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_color(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_faces(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_edges(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_vertices(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_volume(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_bounding_box(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def set_assigned_material(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_assigned_material(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def remove_assigned_material(self, **kwargs):  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def set_name(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def set_fill_style(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def set_suppressed(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def set_color(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def rotate(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def scale(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def mirror(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def map(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_collision(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def copy(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_tesellation(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def get_tesellation_with_options(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError

    @protect_grpc
    def boolean(self, **kwargs) -> dict:  # noqa: D102
        raise NotImplementedError



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a box (quadrilateral)."""

from typing import TYPE_CHECKING

from beartype import beartype as check_input_types
from pint import Quantity
from scipy.spatial.transform import Rotation as SpatialRotation

from ansys.geometry.core.math.matrix import Matrix33
from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.checks import graphics_required
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.segment import SketchSegment
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv


class Box(SketchFace):
    """Provides for modeling a box.

    Parameters
    ----------
    center: Point2D
        Center point of the box.
    width : ~pint.Quantity | Distance | Real
        Width of the box.
    height : ~pint.Quantity | Distance | Real
        Height of the box.
    angle : ~pint.Quantity | Angle | Real, default: 0
        Placement angle for orientation alignment.
    """

    @check_input_types
    def __init__(
        self,
        center: Point2D,
        width: Quantity | Distance | Real,
        height: Quantity | Distance | Real,
        angle: Quantity | Angle | Real = 0,
    ):
        """Initialize the box."""
        super().__init__()

        self._center = center
        if isinstance(angle, (int, float)):
            angle = Angle(angle, DEFAULT_UNITS.ANGLE)
        angle = angle if isinstance(angle, Angle) else Angle(angle, angle.units)

        rotation = Matrix33(
            SpatialRotation.from_euler(
                "xyz", [0, 0, angle.value.m_as(UNITS.radian)], degrees=False
            ).as_matrix()
        )

        self._width = width if isinstance(width, Distance) else Distance(width, center.unit)
        if self._width.value <= 0:
            raise ValueError("Width must be a real positive value.")
        width_magnitude = self._width.value.m_as(center.unit)

        self._height = height if isinstance(height, Distance) else Distance(height, center.unit)
        if self._height.value <= 0:
            raise ValueError("Height must be a real positive value.")
        height_magnitude = self._height.value.m_as(center.unit)

        half_h = height_magnitude / 2
        half_w = width_magnitude / 2
        corner_1 = rotation @ [-half_w, half_h, 0]
        corner_2 = rotation @ [half_w, half_h, 0]
        corner_3 = rotation @ [half_w, -half_h, 0]
        corner_4 = rotation @ [-half_w, -half_h, 0]

        self._corner_1 = Point2D([center.x.m + corner_1[0], center.y.m + corner_1[1]], center.unit)
        self._corner_2 = Point2D([center.x.m + corner_2[0], center.y.m + corner_2[1]], center.unit)
        self._corner_3 = Point2D([center.x.m + corner_3[0], center.y.m + corner_3[1]], center.unit)
        self._corner_4 = Point2D([center.x.m + corner_4[0], center.y.m + corner_4[1]], center.unit)

        # TODO: add plane to SketchSegment when available
        # https://github.com/ansys/pyansys-geometry/issues/1319
        self._width_segment1 = SketchSegment(self._corner_1, self._corner_2)
        self._height_segment1 = SketchSegment(self._corner_2, self._corner_3)
        self._width_segment2 = SketchSegment(self._corner_3, self._corner_4)
        self._height_segment2 = SketchSegment(self._corner_4, self._corner_1)

        self._edges.append(self._width_segment1)
        self._edges.append(self._height_segment1)
        self._edges.append(self._width_segment2)
        self._edges.append(self._height_segment2)

    @property
    def center(self) -> Point2D:
        """Center point of the box."""
        return self._center

    @property
    def width(self) -> Quantity:
        """Width of the box."""
        return self._width.value

    @property
    def height(self) -> Quantity:
        """Height of the box."""
        return self._height.value

    @property
    def perimeter(self) -> Quantity:
        """Perimeter of the box."""
        return 2 * self.width + 2 * self.height

    @property
    def area(self) -> Quantity:
        """Area of the box."""
        return self.width * self.height

    @property
    @graphics_required
    def visualization_polydata(self) -> "pv.PolyData":
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        import numpy as np
        import pyvista as pv

        return pv.Quadrilateral(
            np.array(
                [
                    [
                        self._corner_1.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._corner_1.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._corner_2.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._corner_2.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._corner_3.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._corner_3.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                    [
                        self._corner_4.x.m_as(DEFAULT_UNITS.LENGTH),
                        self._corner_4.y.m_as(DEFAULT_UNITS.LENGTH),
                        0,
                    ],
                ],
                dtype=np.float64,
            )
        )



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing designs."""

from enum import Enum, unique
from pathlib import Path
from typing import Union

from beartype import beartype as check_input_types
from google.protobuf.empty_pb2 import Empty
import numpy as np
from pint import Quantity, UndefinedUnitError

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.commands_pb2 import (
    AssignMidSurfaceOffsetTypeRequest,
    AssignMidSurfaceThicknessRequest,
    CreateBeamCircularProfileRequest,
)
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from ansys.geometry.core.connection.backend import BackendType
from ansys.geometry.core.connection.conversions import (
    grpc_curve_to_curve,
    grpc_frame_to_frame,
    grpc_material_to_material,
    grpc_matrix_to_matrix,
    grpc_point_to_point3d,
    plane_to_grpc_plane,
    point3d_to_grpc_point,
)
from ansys.geometry.core.designer.beam import (
    Beam,
    BeamCircularProfile,
    BeamCrossSectionInfo,
    BeamProfile,
    BeamProperties,
    SectionAnchorType,
)
from ansys.geometry.core.designer.body import Body, MasterBody, MidSurfaceOffsetType
from ansys.geometry.core.designer.component import Component, SharedTopologyType
from ansys.geometry.core.designer.coordinate_system import CoordinateSystem
from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.designer.edge import Edge
from ansys.geometry.core.designer.face import Face
from ansys.geometry.core.designer.part import MasterComponent, Part
from ansys.geometry.core.designer.selection import NamedSelection
from ansys.geometry.core.designer.vertex import Vertex
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.materials.material import Material
from ansys.geometry.core.materials.property import MaterialProperty, MaterialPropertyType
from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Y, ZERO_POINT3D
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.checks import ensure_design_is_active, min_backend_version
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Distance
from ansys.geometry.core.misc.options import ImportOptions
from ansys.geometry.core.modeler import Modeler
from ansys.geometry.core.parameters.parameter import Parameter, ParameterUpdateStatus
from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
from ansys.geometry.core.shapes.parameterization import Interval, ParamUV
from ansys.geometry.core.typing import RealSequence


@unique
class DesignFileFormat(Enum):
    """Provides supported file formats that can be downloaded for designs."""

    SCDOCX = "SCDOCX"
    PARASOLID_TEXT = "PARASOLID_TEXT"
    PARASOLID_BIN = "PARASOLID_BIN"
    FMD = "FMD"
    STEP = "STEP"
    IGES = "IGES"
    PMDB = "PMDB"
    STRIDE = "STRIDE"
    DISCO = "DISCO"
    INVALID = "INVALID"

    def __str__(self):
        """Represent object in string format."""
        return self.value


class Design(Component):
    """Provides for organizing geometry assemblies.

    This class synchronizes to a supporting Geometry service instance.

    Parameters
    ----------
    name : str
        User-defined label for the design.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    read_existing_design : bool, default: False
        Whether an existing design on the service should be read. This parameter is
        only valid when connecting to an existing service session. Otherwise, avoid
        using this optional parameter.
    """

    # Types of the class instance private attributes
    _materials: list[Material]
    _named_selections: dict[str, NamedSelection]
    _beam_profiles: dict[str, BeamProfile]

    @protect_grpc
    @check_input_types
    def __init__(self, name: str, modeler: Modeler, read_existing_design: bool = False):
        """Initialize the ``Design`` class."""
        super().__init__(name, None, modeler.client)

        # Initialize the stubs needed
        self._commands_stub = CommandsStub(self._grpc_client.channel)

        # Initialize needed instance variables
        self._materials = []
        self._named_selections = {}
        self._beam_profiles = {}
        self._design_id = ""
        self._is_active = False
        self._modeler = modeler

        # Check whether we want to process an existing design or create a new one.
        if read_existing_design:
            self._grpc_client.log.debug("Reading Design object from service.")
            self.__read_existing_design()
        else:
            response = self._grpc_client.services.designs.new(name=name)
            self._design_id = response.get("design_id")
            self._id = response.get("main_part_id")
            self._activate(called_after_design_creation=True)
            self._grpc_client.log.debug("Design object instantiated successfully.")

    @property
    def design_id(self) -> str:
        """The design's object unique id."""
        return self._design_id

    @property
    def materials(self) -> list[Material]:
        """List of materials available for the design."""
        return self._materials

    @property
    def named_selections(self) -> list[NamedSelection]:
        """List of named selections available for the design."""
        return list(self._named_selections.values())

    @property
    def beam_profiles(self) -> list[BeamProfile]:
        """List of beam profile available for the design."""
        return list(self._beam_profiles.values())

    @property
    def parameters(self) -> list[Parameter]:
        """List of parameters available for the design."""
        return self.get_all_parameters()

    @property
    def is_active(self) -> bool:
        """Whether the design is currently active."""
        return self._is_active

    @property
    def is_closed(self) -> bool:
        """Whether the design is closed (i.e. not active)."""
        return not self._is_active

    def close(self) -> None:
        """Close the design."""
        # Check if the design is already closed
        if self.is_closed:
            self._grpc_client.log.warning(f"Design {self.name} is already closed.")
            return

        # Attempt to close the design
        try:
            self._grpc_client.services.designs.close(design_id=self._design_id)
        except Exception as err:
            self._grpc_client.log.warning(f"Design {self.name} could not be closed. Error: {err}.")
            self._grpc_client.log.warning("Ignoring response and assuming the design is closed.")

        # Consider the design closed (even if the close request failed)
        self._is_active = False

    def _activate(self, called_after_design_creation: bool = False) -> None:
        """Activate the design."""
        # Activate the current design
        if not called_after_design_creation:
            self._grpc_client.services.designs.put_active(design_id=self._design_id)
        self._is_active = True
        self._grpc_client.log.debug(f"Design {self.name} is activated.")

    # TODO: allow for list of materials
    # https://github.com/ansys/pyansys-geometry/issues/1319
    @check_input_types
    @ensure_design_is_active
    def add_material(self, material: Material) -> None:
        """Add a material to the design.

        Parameters
        ----------
        material : Material
            Material to add.
        """
        self._grpc_client.services.materials.add_material(material=material)
        self._materials.append(material)
        self._grpc_client.log.debug(f"Material {material.name} is successfully added to design.")

    @min_backend_version(26, 1, 0)
    @check_input_types
    @ensure_design_is_active
    def remove_material(self, material: Material | list[Material]) -> None:
        """Remove a material from the design.

        Parameters
        ----------
        material : Material | list[Material]
            Material or list of materials to remove.
        """
        material = material if isinstance(material, list) else [material]

        self._grpc_client.services.materials.remove_material(materials=material)
        for mat in material:
            self._materials.remove(mat)
            self._grpc_client.log.debug(f"Material {mat.name} is successfully removed from design.")

    @check_input_types
    @ensure_design_is_active
    def save(self, file_location: Path | str, write_body_facets: bool = False) -> None:
        """Save a design to disk on the active Geometry server instance.

        Parameters
        ----------
        file_location : ~pathlib.Path | str
            Location on disk to save the file to.
        write_body_facets : bool, default: False
            Option to write body facets into the saved file. 26R1 and later.
        """
        # Sanity checks on inputs
        if isinstance(file_location, Path):
            file_location = str(file_location)

        self._grpc_client.services.designs.save_as(
            filepath=file_location,
            write_body_facets=write_body_facets,
            backend_version=self._grpc_client.backend_version,
        )
        self._grpc_client.log.debug(f"Design successfully saved at location {file_location}.")

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def download(
        self,
        file_location: Path | str,
        format: DesignFileFormat = DesignFileFormat.SCDOCX,
        write_body_facets: bool = False,
    ) -> None:
        """Export and download the design from the server.

        Parameters
        ----------
        file_location : ~pathlib.Path | str
            Location on disk to save the file to.
        format : DesignFileFormat, default: DesignFileFormat.SCDOCX
            Format for the file to save to.
        write_body_facets : bool, default: False
            Option to write body facets into the saved file. SCDOCX only, 26R1 and later.
        """
        # Sanity checks on inputs
        if isinstance(file_location, str):
            file_location = Path(file_location)

        # Check if the folder for the file location exists
        if not file_location.parent.exists():
            # Create the parent directory
            file_location.parent.mkdir(parents=True, exist_ok=True)

        # Process response
        self._grpc_client.log.debug(f"Requesting design download in {format} format.")
        if self._modeler.client.backend_version < (25, 2, 0):
            received_bytes = self.__export_and_download_legacy(format=format)
        else:
            received_bytes = self.__export_and_download(
                format=format, write_body_facets=write_body_facets
            )

        # Write to file
        file_location.write_bytes(received_bytes)
        self._grpc_client.log.debug(f"Design downloaded at location {file_location}.")

    def __export_and_download_legacy(self, format: DesignFileFormat) -> bytes:
        """Export and download the design from the server.

        Parameters
        ----------
        format : DesignFileFormat
            Format for the file to save to.

        Returns
        -------
        bytes
            The raw data from the exported and downloaded file.

        Notes
        -----
        This is a legacy method, which is used in versions
        up to Ansys 25.1.1 products.
        """
        # Process response
        self._grpc_client.log.debug(f"Requesting design download in {format} format.")
        if format is DesignFileFormat.SCDOCX:
            response = self._commands_stub.DownloadFile(Empty())
            received_bytes = bytes()
            received_bytes += response.data
        elif format in [
            DesignFileFormat.PARASOLID_TEXT,
            DesignFileFormat.PARASOLID_BIN,
            DesignFileFormat.FMD,
            DesignFileFormat.STEP,
            DesignFileFormat.IGES,
            DesignFileFormat.PMDB,
        ]:
            response = self._grpc_client.services.parts.export(format=format)
            received_bytes = response.get("data")
        else:
            self._grpc_client.log.warning(
                f"{format} format requested is not supported. Ignoring download request."
            )
            return

        return received_bytes

    def __export_and_download(
        self,
        format: DesignFileFormat,
        write_body_facets: bool = False,
    ) -> bytes:
        """Export and download the design from the server.

        Parameters
        ----------
        format : DesignFileFormat
            Format for the file to save to.

        Returns
        -------
        bytes
            The raw data from the exported and downloaded file.
        """
        # Process response
        self._grpc_client.log.debug(f"Requesting design download in {format} format.")

        if format in [
            DesignFileFormat.PARASOLID_TEXT,
            DesignFileFormat.PARASOLID_BIN,
            DesignFileFormat.FMD,
            DesignFileFormat.STEP,
            DesignFileFormat.IGES,
            DesignFileFormat.PMDB,
            DesignFileFormat.DISCO,
            DesignFileFormat.SCDOCX,
            DesignFileFormat.STRIDE,
        ]:
            try:
                response = self._grpc_client.services.designs.download_export(
                    format=format,
                    write_body_facets=write_body_facets,
                    backend_version=self._grpc_client.backend_version,
                )
            except Exception:
                self._grpc_client.log.warning(
                    f"Failed to download the file in {format} format."
                    " Attempting to stream download."
                )
                # Attempt to download the file via streaming
                response = self._grpc_client.services.designs.stream_download_export(
                    format=format,
                    write_body_facets=write_body_facets,
                    backend_version=self._grpc_client.backend_version,
                )
        else:
            self._grpc_client.log.warning(
                f"{format} format requested is not supported. Ignoring download request."
            )
            return None

        return response.get("data")

    def __build_export_file_location(self, location: Path | str | None, ext: str) -> Path:
        """Build the file location for export functions.

        Parameters
        ----------
        location : ~pathlib.Path | str
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.
        ext : str
            Extension to use for the file.

        Returns
        -------
        ~pathlib.Path
            The file location for the export function.
        """
        return (Path(location) if location else Path.cwd()) / f"{self.name}.{ext}"

    def export_to_scdocx(self, location: Path | str | None = None) -> Path:
        """Export the design to an scdocx file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "scdocx")

        # Export the design to an scdocx file
        self.download(file_location, DesignFileFormat.SCDOCX)

        # Return the file location
        return file_location

    def export_to_disco(self, location: Path | str | None = None) -> Path:
        """Export the design to an dsco file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "dsco")

        # Export the design to an dsco file
        self.download(file_location, DesignFileFormat.DISCO)

        # Return the file location
        return file_location

    def export_to_stride(self, location: Path | str | None = None) -> Path:
        """Export the design to an stride file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "stride")

        # Export the design to an stride file
        self.download(file_location, DesignFileFormat.STRIDE)

        # Return the file location
        return file_location

    def export_to_parasolid_text(self, location: Path | str | None = None) -> Path:
        """Export the design to a Parasolid text file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Determine the extension based on the backend type
        ext = "xmt_txt" if BackendType.is_linux_service(self._grpc_client.backend_type) else "x_t"

        # Define the file location
        file_location = self.__build_export_file_location(location, ext)

        # Export the design to a Parasolid text file
        self.download(file_location, DesignFileFormat.PARASOLID_TEXT)

        # Return the file location
        return file_location

    def export_to_parasolid_bin(self, location: Path | str | None = None) -> Path:
        """Export the design to a Parasolid binary file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Determine the extension based on the backend type
        ext = "xmt_bin" if BackendType.is_linux_service(self._grpc_client.backend_type) else "x_b"

        # Define the file location
        file_location = self.__build_export_file_location(location, ext)

        # Export the design to a Parasolid binary file
        self.download(file_location, DesignFileFormat.PARASOLID_BIN)

        # Return the file location
        return file_location

    def export_to_fmd(self, location: Path | str | None = None) -> Path:
        """Export the design to an FMD file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "fmd")

        # Export the design to an FMD file
        self.download(file_location, DesignFileFormat.FMD)

        # Return the file location
        return file_location

    def export_to_step(self, location: Path | str | None = None) -> Path:
        """Export the design to a STEP file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "stp")

        # Export the design to a STEP file
        self.download(file_location, DesignFileFormat.STEP)

        # Return the file location
        return file_location

    def export_to_iges(self, location: Path | str = None) -> Path:
        """Export the design to an IGES file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "igs")

        # Export the design to an IGES file
        self.download(file_location, DesignFileFormat.IGES)

        # Return the file location
        return file_location

    def export_to_pmdb(self, location: Path | str | None = None) -> Path:
        """Export the design to a PMDB file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "pmdb")

        # Export the design to a PMDB file
        self.download(file_location, DesignFileFormat.PMDB)

        # Return the file location
        return file_location

    @check_input_types
    @ensure_design_is_active
    def create_named_selection(
        self,
        name: str,
        bodies: list[Body] | None = None,
        faces: list[Face] | None = None,
        edges: list[Edge] | None = None,
        beams: list[Beam] | None = None,
        design_points: list[DesignPoint] | None = None,
        components: list[Component] | None = None,
        vertices: list[Vertex] | None = None,
    ) -> NamedSelection:
        """Create a named selection on the active Geometry server instance.

        Parameters
        ----------
        name : str
            User-defined name for the named selection.
        bodies : list[Body], default: None
            All bodies to include in the named selection.
        faces : list[Face], default: None
            All faces to include in the named selection.
        edges : list[Edge], default: None
            All edges to include in the named selection.
        beams : list[Beam], default: None
            All beams to include in the named selection.
        design_points : list[DesignPoint], default: None
            All design points to include in the named selection.
        components : list[Component], default: None
            All components to include in the named selection.
        vertices : list[Vertex], default: None
            All vertices to include in the named selection.

        Returns
        -------
        NamedSelection
            Newly created named selection that maintains references to all target entities.

        Raises
        ------
        ValueError
            If no entities are provided for the named selection. At least
            one of the optional parameters must be provided.
        """
        # Verify that at least one entity is provided
        if not any([bodies, faces, edges, beams, design_points, components, vertices]):
            raise ValueError(
                "At least one of the following must be provided: "
                "bodies, faces, edges, beams, design_points, components, or vertices."
            )

        named_selection = NamedSelection(
            name,
            self,
            self._grpc_client,
            bodies=bodies,
            faces=faces,
            edges=edges,
            beams=beams,
            design_points=design_points,
            components=components,
            vertices=vertices,
        )

        self._named_selections[named_selection.name] = named_selection
        self._grpc_client.log.debug(
            f"Named selection {named_selection.name} is successfully created."
        )

        return self._named_selections[named_selection.name]

    @check_input_types
    @ensure_design_is_active
    def delete_named_selection(self, named_selection: NamedSelection | str) -> None:
        """Delete a named selection on the active Geometry server instance.

        Parameters
        ----------
        named_selection : NamedSelection | str
            Name of the named selection or instance.
        """
        if isinstance(named_selection, str):
            removal_name = named_selection
            removal = self._named_selections.get(named_selection, None)
            removal_id = removal.id if removal else None
        else:
            removal_name = named_selection.name
            removal_id = named_selection.id

        self._grpc_client.log.debug(f"Named selection {removal_name} deletion request received.")
        self._grpc_client.services.named_selection.delete_named_selection(id=removal_id)

        try:
            self._named_selections.pop(removal_name)
            self._grpc_client.log.debug(f"Named selection {removal_name} is successfully deleted.")
        except KeyError:
            self._grpc_client.log.warning(
                f"Attempted named selection deletion failed, with name {removal_name}."
                + " Ignoring this request."
            )
            pass

    @check_input_types
    @ensure_design_is_active
    def delete_component(self, component: Union["Component", str]) -> None:
        """Delete a component (itself or its children).

        Parameters
        ----------
        id : Union[Component, str]
            Name of the component or instance to delete.

        Raises
        ------
        ValueError
            The design itself cannot be deleted.

        Notes
        -----
        If the component is not this component (or its children), it
        is not deleted.
        """
        id = component if isinstance(component, str) else component.id
        if id == self.id:
            raise ValueError("The design itself cannot be deleted.")
        else:
            return super().delete_component(component)

    def set_shared_topology(self, share_type: SharedTopologyType) -> None:
        """Set the shared topology to apply to the component.

        Parameters
        ----------
        share_type : SharedTopologyType
            Shared topology type to assign.

        Raises
        ------
        ValueError
            Shared topology does not apply to a design.
        """
        raise ValueError("The design itself cannot have a shared topology.")

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def add_beam_circular_profile(
        self,
        name: str,
        radius: Quantity | Distance,
        center: np.ndarray | RealSequence | Point3D = ZERO_POINT3D,
        direction_x: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_X,
        direction_y: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_Y,
    ) -> BeamCircularProfile:
        """Add a new beam circular profile under the design for creating beams.

        Parameters
        ----------
        name : str
            User-defined label for the new beam circular profile.
        radius : ~pint.Quantity | Distance
            Radius of the beam circular profile.
        center : ~numpy.ndarray | RealSequence | Point3D
            Center of the beam circular profile.
        direction_x : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
            X-plane direction.
        direction_y : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
            Y-plane direction.
        """
        dir_x = direction_x if isinstance(direction_x, UnitVector3D) else UnitVector3D(direction_x)
        dir_y = direction_y if isinstance(direction_y, UnitVector3D) else UnitVector3D(direction_y)
        radius = radius if isinstance(radius, Distance) else Distance(radius)

        if radius.value <= 0:
            raise ValueError("Radius must be a real positive value.")

        if not dir_x.is_perpendicular_to(dir_y):
            raise ValueError("Direction X and direction Y must be perpendicular.")

        request = CreateBeamCircularProfileRequest(
            origin=point3d_to_grpc_point(center),
            radius=radius.value.m_as(DEFAULT_UNITS.SERVER_LENGTH),
            plane=plane_to_grpc_plane(Plane(center, dir_x, dir_y)),
            name=name,
        )

        self._grpc_client.log.debug(f"Creating a beam circular profile on {self.id}...")

        response = self._commands_stub.CreateBeamCircularProfile(request)
        profile = BeamCircularProfile(response.id, name, radius, center, dir_x, dir_y)
        self._beam_profiles[profile.name] = profile

        self._grpc_client.log.debug(
            f"Beam circular profile {profile.name} is successfully created."
        )

        return self._beam_profiles[profile.name]

    @min_backend_version(25, 1, 0)
    def get_all_parameters(self) -> list[Parameter]:
        """Get parameters for the design.

        Returns
        -------
        list[Parameter]
            List of parameters for the design.

        Warnings
        --------
        This method is only available starting on Ansys release 25R1.
        """
        response = self._grpc_client._services.driving_dimensions.get_all_parameters()
        return response.get("parameters")

    @check_input_types
    @min_backend_version(25, 1, 0)
    def set_parameter(self, dimension: Parameter) -> ParameterUpdateStatus:
        """Set or update a parameter of the design.

        Parameters
        ----------
        dimension : Parameter
            Parameter to set.

        Returns
        -------
        ParameterUpdateStatus
            Status of the update operation.

        Warnings
        --------
        This method is only available starting on Ansys release 25R1.
        """
        response = self._grpc_client._services.driving_dimensions.set_parameter(
            driving_dimension=dimension
        )

        # Update the design in place. This method is computationally expensive,
        # consider finding a more efficient approach.
        self._update_design_inplace()

        return response.get("status")

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def add_midsurface_thickness(self, thickness: Quantity, bodies: list[Body]) -> None:
        """Add a mid-surface thickness to a list of bodies.

        Parameters
        ----------
        thickness : ~pint.Quantity
            Thickness to be assigned.
        bodies : list[Body]
            All bodies to include in the mid-surface thickness assignment.

        Notes
        -----
        Only surface bodies will be eligible for mid-surface thickness assignment.
        """
        # Store only assignable ids
        ids: list[str] = []
        ids_bodies: list[Body] = []
        for body in bodies:
            if body.is_surface:
                ids.append(body.id)
                ids_bodies.append(body)
            else:
                self._grpc_client.log.warning(
                    f"Body {body.name} cannot be assigned a mid-surface thickness since it is not a surface. Ignoring request."  # noqa : E501
                )

        # Assign mid-surface thickness
        self._commands_stub.AssignMidSurfaceThickness(
            AssignMidSurfaceThicknessRequest(
                bodies_or_faces=ids, thickness=thickness.m_as(DEFAULT_UNITS.SERVER_LENGTH)
            )
        )

        # Once the assignment has gone fine, store the values
        for body in ids_bodies:
            body._surface_thickness = thickness

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def add_midsurface_offset(self, offset_type: MidSurfaceOffsetType, bodies: list[Body]) -> None:
        """Add a mid-surface offset type to a list of bodies.

        Parameters
        ----------
        offset_type : MidSurfaceOffsetType
            Surface offset to be assigned.
        bodies : list[Body]
            All bodies to include in the mid-surface offset assignment.

        Notes
        -----
        Only surface bodies will be eligible for mid-surface offset assignment.
        """
        # Store only assignable ids
        ids: list[str] = []
        ids_bodies: list[Body] = []
        for body in bodies:
            if body.is_surface:
                ids.append(body.id)
                ids_bodies.append(body)
            else:
                self._grpc_client.log.warning(
                    f"Body {body.name} cannot be assigned a mid-surface offset since it is not a surface. Ignoring request."  # noqa : E501
                )

        # Assign mid-surface offset type
        self._commands_stub.AssignMidSurfaceOffsetType(
            AssignMidSurfaceOffsetTypeRequest(bodies_or_faces=ids, offset_type=offset_type.value)
        )

        # Once the assignment has gone fine, store the values
        for body in ids_bodies:
            body._surface_offset = offset_type

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def delete_beam_profile(self, beam_profile: BeamProfile | str) -> None:
        """Remove a beam profile on the active geometry server instance.

        Parameters
        ----------
        beam_profile : BeamProfile | str
            A beam profile name or instance that should be deleted.
        """
        removal_name = beam_profile if isinstance(beam_profile, str) else beam_profile.name
        self._grpc_client.log.debug(f"Beam profile {removal_name} deletion request received.")
        removal_obj = self._beam_profiles.get(removal_name, None)

        if removal_obj:
            self._commands_stub.DeleteBeamProfile(EntityIdentifier(id=removal_obj.id))
            self._beam_profiles.pop(removal_name)
            self._grpc_client.log.debug(f"Beam profile {removal_name} successfully deleted.")
        else:
            self._grpc_client.log.warning(
                f"Attempted beam profile deletion failed, with name {removal_name}."
                + " Ignoring request."
            )

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    @min_backend_version(24, 2, 0)
    def insert_file(
        self,
        file_location: Path | str,
        import_options: ImportOptions = ImportOptions(),
    ) -> Component:
        """Insert a file into the design.

        Parameters
        ----------
        file_location : ~pathlib.Path | str
            Location on disk where the file is located.
        import_options : ImportOptions
            The options to pass into upload file

        Returns
        -------
        Component
            The newly inserted component.

        Warnings
        --------
        This method is only available starting on Ansys release 24R2.
        """
        # Upload the file to the server
        filepath_server = self._modeler._upload_file(file_location, import_options=import_options)

        # Insert the file into the design
        self._grpc_client.services.designs.insert(
            filepath=filepath_server, import_named_selections=import_options.import_named_selections
        )
        self._grpc_client.log.debug(f"File {file_location} successfully inserted into design.")

        self._update_design_inplace()

        self._grpc_client.log.debug(f"Design {self.name} is successfully updated.")

        # Return the newly inserted component
        return self._components[-1]

    def __repr__(self) -> str:
        """Represent the ``Design`` as a string."""
        alive_bodies = [1 if body.is_alive else 0 for body in self.bodies]
        alive_comps = [1 if comp.is_alive else 0 for comp in self.components]
        lines = [f"ansys.geometry.core.designer.Design {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Is active?           : {self._is_active}")
        lines.append(f"  N Bodies             : {sum(alive_bodies)}")
        lines.append(f"  N Components         : {sum(alive_comps)}")
        lines.append(f"  N Coordinate Systems : {len(self.coordinate_systems)}")
        lines.append(f"  N Named Selections   : {len(self.named_selections)}")
        lines.append(f"  N Materials          : {len(self.materials)}")
        lines.append(f"  N Beam Profiles      : {len(self.beam_profiles)}")
        lines.append(f"  N Design Points      : {len(self.design_points)}")
        return "\n".join(lines)

    def __read_existing_design(self) -> None:
        """Read an existing ``Design`` located on the server."""
        #
        # This might go out of sync with the _update_design_inplace method.
        # Ensure that the two methods are in sync. Especially regarding cleanup.
        #
        # TODO: Not all features implemented yet. Status is as follows
        #
        # Windows:
        #
        # - [X] Components
        # - [X] Bodies
        # - [X] Materials
        # - [X] NamedSelections
        # - [ ] BeamProfiles
        # - [ ] Beams
        # - [X] CoordinateSystems
        # - [X] SharedTopology
        #
        # Linux:
        #
        # - [X] Components
        # - [X] Bodies
        # - [X] Materials
        # - [X] NamedSelections
        # - [ ] BeamProfiles
        # - [ ] Beams
        # - [X] CoordinateSystems
        # - [ ] SharedTopology
        #
        # https://github.com/ansys/pyansys-geometry/issues/1319
        import time

        start = time.time()
        # Grab active design
        design_response = self._grpc_client.services.designs.get_active()
        if not design_response:
            raise RuntimeError("No existing design available at service level.")
        else:
            self._design_id = design_response.get("design_id")
            self._id = design_response.get("main_part_id")
            self._name = design_response.get("name")
            self._activate(called_after_design_creation=True)

        response = self._commands_stub.GetAssembly(EntityIdentifier(id=self._design_id))

        # Store created objects
        created_parts = {p.id: Part(p.id, p.name, [], []) for p in response.parts}
        created_tps = {}
        created_components = {design_response.get("main_part_id"): self}
        created_bodies = {}

        # Make dummy master for design since server doesn't have one
        self._master_component = MasterComponent(
            "1", "master_design", created_parts[design_response.get("main_part_id")]
        )

        # Create MasterComponents
        for master in response.transformed_parts:
            part = created_parts.get(master.part_master.id)
            new_master = MasterComponent(
                master.id, master.name, part, grpc_matrix_to_matrix(master.placement)
            )
            created_tps[master.id] = new_master

        # Create Components
        for comp in response.components:
            parent = created_components.get(comp.parent_id)
            master = created_tps.get(comp.master_id)
            c = Component(
                comp.name,
                parent,
                self._grpc_client,
                preexisting_id=comp.id,
                master_component=master,
                read_existing_comp=True,
            )
            created_components[comp.id] = c
            parent.components.append(c)

        # Create Bodies
        for body in response.bodies:
            part = created_parts.get(body.parent_id)
            tb = MasterBody(body.id, body.name, self._grpc_client, is_surface=body.is_surface)
            part.bodies.append(tb)
            created_bodies[body.id] = tb

        # Create Materials
        for material in response.materials:
            properties = []
            density = Quantity(0)
            for property in material.material_properties:
                # TODO: Add support for more material properties...
                #      - Need to add support for more MaterialPropertyTypes
                #      - Need to add support for more Quantity units
                # https://github.com/ansys/pyansys-geometry/issues/1319
                try:
                    mp_type = MaterialPropertyType.from_id(property.id)
                except ValueError as err:
                    # TODO: Errors coming from MaterialPropertyType.from_id
                    # because of unsupported MaterialPropertyType entries...
                    # https://github.com/ansys/pyansys-geometry/issues/1319
                    self._grpc_client.log.warning(
                        f"Material property {property.display_name} of type {property.id} is not supported."  # noqa : E501
                        " Storing as string."
                    )
                    self._grpc_client.log.warning(f"Root cause: {err}")
                    mp_type = property.id

                try:
                    mp_quantity = Quantity(property.value, property.units)
                except (
                    UndefinedUnitError,
                    TypeError,
                ) as err:  # TODO: Errors coming from Quantity ctor because of unsupported units...
                    # https://github.com/ansys/pyansys-geometry/issues/1319
                    self._grpc_client.log.warning(
                        f"Material property {property.display_name} with units {property.units} is not fully supported."  # noqa : E501
                        " Storing value only as float."
                    )
                    self._grpc_client.log.warning(f"Root cause: {err}")
                    mp_quantity = property.value

                mp = MaterialProperty(mp_type, property.display_name, mp_quantity)
                properties.append(mp)
                if mp.type == MaterialPropertyType.DENSITY:
                    density = (
                        mp.quantity if isinstance(mp.quantity, Quantity) else Quantity(mp.quantity)
                    )

            m = Material(material.name, density, properties)
            self.materials.append(m)

        # Create Beams
        for beam in response.beams:
            cross_section = BeamCrossSectionInfo(
                section_anchor=SectionAnchorType(beam.cross_section.section_anchor),
                section_angle=beam.cross_section.section_angle,
                section_frame=grpc_frame_to_frame(beam.cross_section.section_frame),
                section_profile=[
                    [
                        TrimmedCurve(
                            geometry=grpc_curve_to_curve(curve.curve),
                            start=grpc_point_to_point3d(curve.start),
                            end=grpc_point_to_point3d(curve.end),
                            interval=Interval(curve.interval_start, curve.interval_end),
                            length=curve.length,
                        )
                        for curve in curve_list.curves
                    ]
                    for curve_list in beam.cross_section.section_profile
                ],
            )
            properties = BeamProperties(
                area=beam.properties.area,
                centroid=ParamUV(beam.properties.centroid_x, beam.properties.centroid_y),
                warping_constant=beam.properties.warping_constant,
                ixx=beam.properties.ixx,
                ixy=beam.properties.ixy,
                iyy=beam.properties.iyy,
                shear_center=ParamUV(
                    beam.properties.shear_center_x, beam.properties.shear_center_y
                ),
                torsion_constant=beam.properties.torsional_constant,
            )

            new_beam = Beam(
                id=beam.id.id,
                start=grpc_point_to_point3d(beam.shape.start),
                end=grpc_point_to_point3d(beam.shape.end),
                profile=None,
                # TODO: Beams need BeamProfiles imported from existing design
                # https://github.com/ansys/pyansys-geometry/issues/1825
                parent_component=self,
                name=beam.name,
                is_deleted=beam.is_deleted,
                is_reversed=beam.is_reversed,
                is_rigid=beam.is_rigid,
                material=grpc_material_to_material(beam.material),
                cross_section=cross_section,
                properties=properties,
                shape=beam.shape,
                beam_type=beam.type,
            )

            # Find the component to which the beam belongs
            parent = created_components.get(beam.parent.id, self)
            parent._beams.append(new_beam)

        # Create NamedSelections
        for ns in response.named_selections:
            new_ns = NamedSelection(
                ns.name,
                self,
                self._grpc_client,
                preexisting_id=ns.id,
            )
            self._named_selections[new_ns.name] = new_ns

        # Create CoordinateSystems
        num_created_coord_systems = 0
        for component_id, coordinate_systems in response.component_coord_systems.items():
            component = created_components.get(component_id)
            for cs in coordinate_systems.coordinate_systems:
                frame = grpc_frame_to_frame(cs.frame)
                new_cs = CoordinateSystem(cs.name, frame, component, self._grpc_client, cs.id)
                component.coordinate_systems.append(new_cs)
                num_created_coord_systems += 1

        end = time.time()

        # Set SharedTopology
        # TODO: Maybe just add it to Component or Part message
        # we're starting to iterate through all the Components too much.
        # Make sure design doesn't need edge case attention
        # https://github.com/ansys/pyansys-geometry/issues/1319
        num_created_shared_topologies = 0
        for component_id, shared_topology_type in response.component_shared_topologies.items():
            component = created_components.get(component_id)
            component._shared_topology = SharedTopologyType(shared_topology_type)
            num_created_shared_topologies += 1

        self._grpc_client.log.debug(f"Parts created: {len(created_parts)}")
        self._grpc_client.log.debug(f"MasterComponents created: {len(created_tps) + 1}")
        self._grpc_client.log.debug(f"Components created: {len(created_components)}")
        self._grpc_client.log.debug(f"Bodies created: {len(created_bodies)}")
        self._grpc_client.log.debug(f"Materials created: {len(self.materials)}")
        self._grpc_client.log.debug(f"NamedSelections created: {len(self.named_selections)}")
        self._grpc_client.log.debug(f"CoordinateSystems created: {num_created_coord_systems}")
        self._grpc_client.log.debug(f"SharedTopologyTypes set: {num_created_shared_topologies}")

        self._grpc_client.log.debug(f"\nSuccessfully read design in: {end - start} s")

    def _update_design_inplace(self) -> None:
        """Update the design to align with the server side.

        Notes
        -----
        This method is used to update the design inside repair tools.
        Its usage is not recommended for other purposes.
        """
        # Clear all the existing information
        #
        # TODO: This might go out of sync with the __read_existing_design method
        # if the latter is updated and this method is not. Ensure that
        # the two methods are in sync.
        # https://github.com/ansys/pyansys-geometry/issues/1319
        #
        self._components = []
        self._clear_cached_bodies()
        self._materials = []
        self._named_selections = {}
        self._coordinate_systems = {}

        # Read the existing design
        self.__read_existing_design()

    def _update_from_tracker(self, tracker_response: list[dict]):
        """Update the design with the changed bodies while preserving unchanged ones."""
        self._grpc_client.log.debug(
            f"Starting _update_from_tracker with response: {tracker_response}"
        )
        self._handle_modified_bodies(tracker_response.get("modified_bodies", []))
        self._handle_deleted_bodies(tracker_response.get("deleted_bodies", []))
        self._handle_created_bodies(tracker_response.get("created_bodies", []))

    def _handle_modified_bodies(self, modified_bodies):
        for body_info in modified_bodies:
            body_id = body_info["id"]
            body_name = body_info["name"]
            self._grpc_client.log.debug(
                f"Processing modified body: ID={body_id}, Name='{body_name}'"
            )
            updated = False

            for body in self.bodies:
                if body.id == body_id:
                    self._update_body(body, body_info)
                    updated = True
                    self._grpc_client.log.debug(
                        f"Modified body '{body_name}' (ID: {body_id}) updated at root level."
                    )
                    break

            if not updated:
                for component in self.components:
                    if self._find_and_update_body(body_info, component):
                        break

    def _handle_deleted_bodies(self, deleted_bodies):
        for body_info in deleted_bodies:
            body_id = body_info["id"]
            self._grpc_client.log.debug(f"Processing deleted body: ID={body_id}")
            removed = False

            for body in self.bodies:
                if body.id == body_id:
                    body._is_alive = False
                    self.bodies.remove(body)
                    removed = True
                    self._grpc_client.log.info(
                        f"Deleted body (ID: {body_id}) removed from root level."
                    )
                    break

            if not removed:
                for component in self.components:
                    if self._find_and_remove_body(body_info, component):
                        break

    def _handle_created_bodies(self, created_bodies):
        for body_info in created_bodies:
            body_id = body_info["id"]
            body_name = body_info["name"]
            is_surface = body_info.get("is_surface", False)
            self._grpc_client.log.debug(
                f"Processing created body: ID={body_id}, Name='{body_name}'"
            )

            if any(body.id == body_id for body in self.bodies):
                self._grpc_client.log.debug(
                    f"Created body '{body_name}' (ID: {body_id}) already exists at root level."
                )
                continue

            added = any(self._find_and_add_body(body_info, self.components))
            if not added:
                new_body = MasterBody(body_id, body_name, self._grpc_client, is_surface=is_surface)
                self.bodies.append(new_body)
                self._grpc_client.log.debug(
                    f"Added new body '{body_name}' (ID: {body_id}) to root level."
                )

    def _update_body(self, existing_body, body_info):
        self._grpc_client.log.debug(
            f"Updating body '{existing_body.name}' "
            f"(ID: {existing_body.id}) with new info: {body_info}"
        )
        existing_body.name = body_info["name"]
        existing_body._template._is_surface = body_info.get("is_surface", False)

    def _find_and_add_body(self, body_info, components):
        for component in components:
            if component.id == body_info["parent_id"]:
                new_body = MasterBody(
                    body_info["id"],
                    body_info["name"],
                    self._grpc_client,
                    is_surface=body_info.get("is_surface", False),
                )
                component.bodies.append(new_body)
                self._grpc_client.log.debug(
                    f"Added new body '{new_body.name}' (ID: {new_body.id}) "
                    f"to component '{component.name}' (ID: {component.id})"
                )
                return True

            if self._find_and_add_body(body_info, component.components):
                return True

        return False

    def _find_and_update_body(self, body_info, component):
        for body in component.bodies:
            if body.id == body_info["id"]:
                self._update_body(body, body_info)
                self._grpc_client.log.debug(
                    f"Updated body '{body.name}' (ID: {body.id}) in component "
                    f"'{component.name}' (ID: {component.id})"
                )
                return True

        for subcomponent in component.components:
            if self._find_and_update_body(body_info, subcomponent):
                return True

        return False

    def _find_and_remove_body(self, body_info, component):
        for body in component.bodies:
            if body.id == body_info["id"]:
                body._is_alive = False
                component.bodies.remove(body)
                self._grpc_client.log.debug(
                    f"Removed body '{body_info['name']}' (ID: {body_info['id']}) from component "
                    f"'{component.name}' (ID: {component.id})"
                )
                return True

        for subcomponent in component.components:
            if self._find_and_remove_body(body_info, subcomponent):
                return True

        return False



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for managing a vertex."""

import numpy as np

from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.typing import RealSequence


class Vertex(Point3D):
    """Represents a single vertex of a body within the design assembly.

    This class synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        The unique identifier for the vertex.
    position : np.ndarray or RealSequence
        The position of the vertex in 3D space.
    """

    def __new__(
        cls,
        id: str,
        position: np.ndarray | RealSequence,
    ):
        """Initialize ``Vertex`` class."""
        # Only pass position and unit to Point3D.__new__
        obj = super().__new__(cls, position)
        return obj

    def __init__(
        self,
        id: str,
        position: np.ndarray | RealSequence,
    ):
        """Initialize the Vertex with a unique identifier."""
        self._id = id
        super().__init__(position)

        # Make immutable
        self.flags.writeable = False

    @property
    def id(self) -> str:
        """Get the unique identifier of the vertex."""
        return self._id

    def __repr__(self) -> str:
        """Return a string representation of the vertex."""
        lines = [f"ansys.geometry.core.designer.Vertex {hex(id(self))}"]
        lines.append(f"  Id                   : {self.id}")
        lines.append(f"  Position                 : {self.position}")
        return "\n".join(lines)
